from typing import List

import torch
from torch import Tensor

from oml.interfaces.miners import ITripletsMinerInBatch, TTripletsIds
from oml.utils.misc_torch import pairwise_dist


class HardTripletsMiner(ITripletsMinerInBatch):
    """
    This miner selects the hardest triplets based on the distances between the features:

    - The hardest `positive` sample has the `maximal` distance to the anchor sample

    - The hardest `negative` sample has the `minimal` distance to the anchor sample

    """

    def _sample(self, features: Tensor, labels: List[int]) -> TTripletsIds:
        """
        This method samples the hardest triplets inside the batch.

        Args:
            features: Features with the shape of ``[batch_size, feature_size]``
            labels: Labels with the size of ``batch_size``

        Returns:
            The batch of the triplets in the order below:
            ``(anchor, positive, negative)``

        """
        assert features.shape[0] == len(labels)

        features = features.clone().detach()

        dist_mat = pairwise_dist(x1=features, x2=features, p=2)

        ids_anchor, ids_pos, ids_neg = self._sample_from_distmat(distmat=dist_mat, labels=labels)

        return ids_anchor, ids_pos, ids_neg

    @staticmethod
    def _sample_from_distmat(distmat: Tensor, labels: List[int]) -> TTripletsIds:
        """
        This method samples the hardest triplets based on the given
        distances matrix. It chooses each sample in the batch as an
        anchor and then finds the hardest positive and negative pair.

        Args:
            distmat: Matrix with the shape of ``[batch_size, batch_size]``
            labels: Labels with the size of ``batch_size``

        Returns:
            The batch of the triplets with the size of ``batch_size``, order is the following:
                ``(anchor, positive, negative)``

        """

        labels = torch.tensor(labels)

        # Ensure labels are a torch tensor
        labels = labels.to(distmat.device)

        batch_size = labels.size(0)

        label_equal = labels.unsqueeze(0) == labels.unsqueeze(1)  # Shape [batch_size, batch_size]
        label_not_equal = ~label_equal

        # Get the hardest positives: argmax over the distance matrix where labels match (i.e. hardest positive)
        dist_pos = distmat.clone()
        dist_pos[label_not_equal] = -float("inf")  # Set non-positives to -inf
        hardest_pos_idx = torch.argmax(dist_pos, dim=1)

        # Get the hardest negatives: argmin over the distance matrix where labels don't match (i.e. hardest negative)
        dist_neg = distmat.clone()
        dist_neg[label_equal] = float("inf")  # Set non-negatives to +inf
        hardest_neg_idx = torch.argmin(dist_neg, dim=1)

        # Return anchor indices, positive indices, and negative indices
        ids_anchor = list(range(batch_size))
        ids_pos = hardest_pos_idx
        ids_neg = hardest_neg_idx

        return ids_anchor, ids_pos.cpu().tolist(), ids_neg.cpu().tolist()


__all__ = ["HardTripletsMiner"]



# isort: skip_file
# fmt: off
# mypy: disable-error-code="no-untyped-def"
# flake8: noqa

# The model is taken from the following project:
#   https://github.com/TaoRuijie/ECAPA-TDNN/blob/main/model.py

'''
This is the ECAPA-TDNN model.
This model is modified and combined based on the following three projects:
  1. https://github.com/clovaai/voxceleb_trainer/issues/86
  2. https://github.com/lawlict/ECAPA-TDNN/blob/master/ecapa_tdnn.py
  3. https://github.com/speechbrain/speechbrain/blob/96077e9a1afff89d3f5ff47cab4bca0202770e4f/speechbrain/lobes/models/ECAPA_TDNN.py

'''

import math

import torch
import torch.nn as nn
import torch.nn.functional as F


class SEModule(nn.Module):
    def __init__(self, channels, bottleneck=128):
        super(SEModule, self).__init__()
        self.se = nn.Sequential(
            nn.AdaptiveAvgPool1d(1),
            nn.Conv1d(channels, bottleneck, kernel_size=1, padding=0),
            nn.ReLU(),
            # nn.BatchNorm1d(bottleneck), # I remove this layer
            nn.Conv1d(bottleneck, channels, kernel_size=1, padding=0),
            nn.Sigmoid(),
            )

    def forward(self, input):
        x = self.se(input)
        return input * x

class Bottle2neck(nn.Module):

    def __init__(self, inplanes, planes, kernel_size=None, dilation=None, scale = 8):
        super(Bottle2neck, self).__init__()
        width       = int(math.floor(planes / scale))
        self.conv1  = nn.Conv1d(inplanes, width*scale, kernel_size=1)
        self.bn1    = nn.BatchNorm1d(width*scale)
        self.nums   = scale -1
        convs       = []
        bns         = []
        num_pad = math.floor(kernel_size/2)*dilation
        for i in range(self.nums):
            convs.append(nn.Conv1d(width, width, kernel_size=kernel_size, dilation=dilation, padding=num_pad))
            bns.append(nn.BatchNorm1d(width))
        self.convs  = nn.ModuleList(convs)
        self.bns    = nn.ModuleList(bns)
        self.conv3  = nn.Conv1d(width*scale, planes, kernel_size=1)
        self.bn3    = nn.BatchNorm1d(planes)
        self.relu   = nn.ReLU()
        self.width  = width
        self.se     = SEModule(planes)

    def forward(self, x):
        residual = x
        out = self.conv1(x)
        out = self.relu(out)
        out = self.bn1(out)

        spx = torch.split(out, self.width, 1)
        for i in range(self.nums):
          if i==0:
            sp = spx[i]
          else:
            sp = sp + spx[i]
          sp = self.convs[i](sp)
          sp = self.relu(sp)
          sp = self.bns[i](sp)
          if i==0:
            out = sp
          else:
            out = torch.cat((out, sp), 1)
        out = torch.cat((out, spx[self.nums]),1)

        out = self.conv3(out)
        out = self.relu(out)
        out = self.bn3(out)

        out = self.se(out)
        out += residual
        return out

class PreEmphasis(torch.nn.Module):

    def __init__(self, coef: float = 0.97):
        super().__init__()
        self.coef = coef
        self.register_buffer(
            'flipped_filter', torch.FloatTensor([-self.coef, 1.]).unsqueeze(0).unsqueeze(0)
        )

    def forward(self, input: torch.tensor) -> torch.tensor:
        input = input.unsqueeze(1)
        input = F.pad(input, (1, 0), 'reflect')
        return F.conv1d(input, self.flipped_filter).squeeze(1)

class FbankAug(nn.Module):

    def __init__(self, freq_mask_width = (0, 8), time_mask_width = (0, 10)):
        self.time_mask_width = time_mask_width
        self.freq_mask_width = freq_mask_width
        super().__init__()

    def mask_along_axis(self, x, dim):
        original_size = x.shape
        batch, fea, time = x.shape
        if dim == 1:
            D = fea
            width_range = self.freq_mask_width
        else:
            D = time
            width_range = self.time_mask_width

        mask_len = torch.randint(width_range[0], width_range[1], (batch, 1), device=x.device).unsqueeze(2)
        mask_pos = torch.randint(0, max(1, D - mask_len.max()), (batch, 1), device=x.device).unsqueeze(2)
        arange = torch.arange(D, device=x.device).view(1, 1, -1)
        mask = (mask_pos <= arange) * (arange < (mask_pos + mask_len))
        mask = mask.any(dim=1)

        if dim == 1:
            mask = mask.unsqueeze(2)
        else:
            mask = mask.unsqueeze(1)

        x = x.masked_fill_(mask, 0.0)
        return x.view(*original_size)

    def forward(self, x):
        x = self.mask_along_axis(x, dim=2)
        x = self.mask_along_axis(x, dim=1)
        return x

class ECAPA_TDNN(nn.Module):

    def __init__(self, C):

        import torchaudio

        super(ECAPA_TDNN, self).__init__()

        self.torchfbank = torch.nn.Sequential(
            PreEmphasis(),
            torchaudio.transforms.MelSpectrogram(sample_rate=16000, n_fft=512, win_length=400, hop_length=160, \
                                                 f_min = 20, f_max = 7600, window_fn=torch.hamming_window, n_mels=80),
            )

        self.specaug = FbankAug() # Spec augmentation

        self.conv1  = nn.Conv1d(80, C, kernel_size=5, stride=1, padding=2)
        self.relu   = nn.ReLU()
        self.bn1    = nn.BatchNorm1d(C)
        self.layer1 = Bottle2neck(C, C, kernel_size=3, dilation=2, scale=8)
        self.layer2 = Bottle2neck(C, C, kernel_size=3, dilation=3, scale=8)
        self.layer3 = Bottle2neck(C, C, kernel_size=3, dilation=4, scale=8)
        # I fixed the shape of the output from MFA layer, that is close to the setting from ECAPA paper.
        self.layer4 = nn.Conv1d(3*C, 1536, kernel_size=1)
        self.attention = nn.Sequential(
            nn.Conv1d(4608, 256, kernel_size=1),
            nn.ReLU(),
            nn.BatchNorm1d(256),
            nn.Tanh(), # I add this layer
            nn.Conv1d(256, 1536, kernel_size=1),
            nn.Softmax(dim=2),
            )
        self.bn5 = nn.BatchNorm1d(3072)
        self.fc6 = nn.Linear(3072, 192)
        self.bn6 = nn.BatchNorm1d(192)


    def forward(self, x, aug):
        with torch.no_grad():
            x = self.torchfbank(x)+1e-6
            x = x.log()
            x = x - torch.mean(x, dim=-1, keepdim=True)
            if aug == True:
                x = self.specaug(x)

        x = self.conv1(x)
        x = self.relu(x)
        x = self.bn1(x)

        x1 = self.layer1(x)
        x2 = self.layer2(x+x1)
        x3 = self.layer3(x+x1+x2)

        x = self.layer4(torch.cat((x1,x2,x3),dim=1))
        x = self.relu(x)

        t = x.size()[-1]

        global_x = torch.cat((x,torch.mean(x,dim=2,keepdim=True).repeat(1,1,t), torch.sqrt(torch.var(x,dim=2,keepdim=True).clamp(min=1e-4)).repeat(1,1,t)), dim=1)

        w = self.attention(global_x)

        mu = torch.sum(x * w, dim=2)
        sg = torch.sqrt( ( torch.sum((x**2) * w, dim=2) - mu**2 ).clamp(min=1e-4) )

        x = torch.cat((mu,sg),1)
        x = self.bn5(x)
        x = self.fc6(x)
        x = self.bn6(x)

        return x



try:
    import pytorch_lightning
except ImportError as e:
    raise ImportError(
        f"{e}\n OML doesn't have lightning as a requirement." f"Run <pip install open-metric-learning[pipelines]>"
    )

from oml.lightning.callbacks.metric import MetricValCallback
from oml.lightning.modules.extractor import ExtractorModule, ExtractorModuleDDP
from oml.lightning.modules.pairwise_postprocessing import (
    PairwiseModule,
    PairwiseModuleDDP,
)
from oml.lightning.pipelines import logging



import inspect
import os
import random
from contextlib import contextmanager
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterable,
    List,
    Sequence,
    Tuple,
    Union,
)

import matplotlib
import matplotlib.pyplot as plt
import numpy as np
import torch
from omegaconf import DictConfig, OmegaConf
from torch import Tensor

from oml.const import BLACK, TCfg, TColor


class CompatibilityError(Exception):
    pass


def adapt_argument_as_kwarg(obj: Callable, candidates: List[str], value: Any) -> Dict[str, Any]:  # type: ignore
    parameters = list(inspect.signature(obj).parameters.keys())

    for name in candidates:
        if name in parameters:
            return {name: value}

    raise CompatibilityError(
        f"Not compatible signature for {obj}. Candidates: {candidates}, available arguments: {parameters}"
    )


def find_value_ids(it: Iterable[Any], value: Any) -> List[int]:
    """
    Args:
        it: List of any
        value: Query element

    Returns:
        Indices of the all elements equal to x0
    """
    if isinstance(it, np.ndarray):
        inds = list(np.where(it == value)[0])
    else:  # could be very slow
        inds = [i for i, el in enumerate(it) if el == value]
    return inds


def set_global_seed(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)

    torch.manual_seed(seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(seed)
        torch.cuda.manual_seed_all(seed)
        torch.backends.cudnn.benchmark = False
        torch.backends.cudnn.deterministic = True

    os.environ["PL_SEED_WORKERS"] = str(1)

    try:
        import torch_xla.core.xla_model as xm
    except ImportError:
        pass
    else:
        xm.set_rng_state(seed)


def one_hot(i: int, dim: int) -> torch.Tensor:
    vector = torch.zeros(dim)
    vector[i] = 1
    return vector


def flatten_dict(
    d: Dict[str, Any], parent_key: str = "", sep: str = "/", ignored_keys: Iterable[str] = ()
) -> Dict[str, Any]:
    items = []  # type: ignore
    for k, v in d.items():
        if k in ignored_keys:
            continue
        new_key = str(parent_key) + sep + str(k) if parent_key else str(k)
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep, ignored_keys=ignored_keys).items())
        else:
            items.append((new_key, v))
    return dict(items)


def dictconfig_to_dict(cfg: TCfg) -> Dict[str, Any]:
    if isinstance(cfg, DictConfig):
        cfg = OmegaConf.to_container(cfg, resolve=True)

    ret = dict()

    for k in cfg.keys():

        if isinstance(cfg[k], DictConfig) or isinstance(cfg[k], dict):
            ret[k] = dictconfig_to_dict(cfg[k])
        else:
            ret[k] = cfg[k]

    return ret


def smart_sample(array: List[Any], k: int) -> List[Any]:
    """Sample n_samples items from given list. If array contains at least n_samples items, sample without repetition;
    otherwise take all the unique items and sample n_samples - len(array) ones with repetition.

    Args:
        array: list of unique elements to sample from
        k: number of items to sample

    Returns:
        sampled_items: list of sampled items
    """
    array_size = len(array)
    if array_size < k:
        sampled = (
            np.random.choice(array, size=array_size, replace=False).tolist()
            + np.random.choice(array, size=k - array_size, replace=True).tolist()
        )
    else:
        sampled = np.random.choice(array, size=k, replace=False).tolist()
    return sampled


def clip_max(arr: Tuple[int, ...], max_el: int) -> Tuple[int, ...]:
    return tuple(min(x, max_el) for x in arr)


def remove_unused_kwargs(kwargs: Dict[str, Any], function: Any) -> Dict[str, Any]:
    return {k: v for k, v in kwargs.items() if k in inspect.signature(function).parameters}


def check_if_nonempty_positive_integers(var: Union[int, Sequence[int]], name: str) -> None:
    """
    Check whether ``var`` is a positive integer or a non-empty Iterable of positive integers.

    Args:
        var: A sequence.
        name: A name of the sequence in case of exception should be raised.

    """
    if isinstance(var, Sequence):
        if not len(var) > 0 or not all([isinstance(x, int) and (x > 0) for x in var]):
            raise ValueError(f"{name} is expected to be non-empty and contain positive integers, but got {var}")
    elif isinstance(var, int):
        if var <= 0:
            raise ValueError(f"{name} is expected to be a positive integer, but got {var}")
    else:
        raise ValueError(f"Unsupported argument type. Expected int or Iterable[int], but got {type(var)}")


def compare_dicts_recursively(d1: Dict, d2: Dict) -> bool:  # type: ignore
    """
    The function compares dictionaries and prints the exact information where they differ. By using the built-in
    dictionary comparison one can get just a plain 'True' or 'False' as result of the comparison, without any hints
    on where the dictionaries differ.
    """
    assert set(d1.keys()) == set(
        d2.keys()
    ), f"The dictionaries keys are different.\nDict_1 keys: {set(d1.keys())}\nDict_2 keys: {set(d2.keys())}"
    for k, v in d1.items():
        if isinstance(v, dict):
            assert compare_dicts_recursively(
                v, d2[k]
            ), f"The dictionaries differs at key {k}.\nDict_1 value: {v}\nDict_2 value: {d2[k]}"
        elif isinstance(v, Tensor):
            assert torch.allclose(d2[k], v)
        else:
            assert d2[k] == v, f"Key name: {k}\nDict_1 value: {v}\nDict_2 value: {d2[k]}"
    return True


def pad_array_right(arr: np.ndarray, required_len: int, val: Union[float, int]) -> np.ndarray:
    assert required_len >= len(arr)
    assert arr.ndim == 1

    return np.pad(arr, (0, required_len - len(arr)), mode="constant", constant_values=val)


def visualize_text(text: str, color: TColor = BLACK, draw_bbox: bool = True) -> np.ndarray:
    with matplotlib_backend("Agg"):
        fig, ax = plt.subplots(figsize=(2.56, 2.56), dpi=100)
        ax.text(0.5, 0.5, text, ha="center", va="center", wrap=True, fontsize=20)
        ax.set_xlim(0, 1)
        ax.set_ylim(0, 1)
        ax.axis("off")

        fig.canvas.draw()
        width, height = fig.canvas.get_width_height()
        image = np.frombuffer(fig.canvas.buffer_rgba(), dtype=np.uint8)
        image = image.reshape(height, width, 4)[..., :3]
        plt.close(fig)

    image = np.resize(image, (256, 256, 3))

    if draw_bbox:
        frame_thickness = 5
        image[:frame_thickness, :, :] = color  # Top frame
        image[-frame_thickness:, :, :] = color  # Bottom frame
        image[:, :frame_thickness, :] = color  # Left frame
        image[:, -frame_thickness:, :] = color  # Right frame

    return image


def html_blocks_to_grid(blocks: List[List[str]]) -> str:
    html_rows = []
    for row in blocks:
        row_html = "<div style='display:flex; justify-content:left;'>" + "".join(row) + "</div>"
        html_rows += [row_html]

    final_html = "<div>" + "\n".join(html_rows) + "</div>"

    return final_html


@contextmanager
def matplotlib_backend(backend: str) -> Generator[None, None, None]:
    current_backend = matplotlib.get_backend()
    try:
        matplotlib.use(backend)
        yield
    finally:
        matplotlib.use(current_backend)


__all__ = [
    "CompatibilityError",
    "adapt_argument_as_kwarg",
    "find_value_ids",
    "set_global_seed",
    "one_hot",
    "flatten_dict",
    "dictconfig_to_dict",
    "smart_sample",
    "clip_max",
    "check_if_nonempty_positive_integers",
    "compare_dicts_recursively",
    "pad_array_right",
    "visualize_text",
    "matplotlib_backend",
    "html_blocks_to_grid",
]



import warnings
from pathlib import Path
from typing import Any, Callable, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import torch
from torch import FloatTensor

from oml.const import (
    AUDIO_EXTENSIONS,
    BLACK,
    DEFAULT_CONVERT_TO_MONO,
    DEFAULT_SAMPLE_RATE,
    INDEX_KEY,
    INPUT_TENSORS_KEY,
    LABELS_KEY,
    MAX_AUDIO_LEN,
    PATHS_COLUMN,
    START_TIME_COLUMN,
    TColor,
)
from oml.datasets.dataframe import (
    DFLabeledDataset,
    DFQueryGalleryDataset,
    DFQueryGalleryLabeledDataset,
)
from oml.interfaces.datasets import (
    IBaseDataset,
    IHTMLVisualizableDataset,
    IVisualizableDataset,
)
from oml.utils.audios import (
    default_spec_repr_func,
    visualize_audio,
    visualize_audio_with_player,
)


def parse_start_times(df: pd.DataFrame) -> Optional[List[float]]:
    """
    Parses starting time points from DataFrame.
    """
    start_times = None
    if START_TIME_COLUMN in df.columns:
        assert (
            df[START_TIME_COLUMN].dtype == float
        ), f"Expected dtype of '{START_TIME_COLUMN}' column is `float`, `{df[START_TIME_COLUMN].dtype}` is found."
        df[START_TIME_COLUMN] = df[START_TIME_COLUMN].fillna(0.0)
        start_times = df[START_TIME_COLUMN].astype(float).tolist()
    return start_times


class AudioBaseDataset(IBaseDataset, IVisualizableDataset, IHTMLVisualizableDataset):
    """
    The base class that handles audio specific logic.
    """

    def __init__(
        self,
        paths: List[Union[str, Path]],
        dataset_root: Optional[Union[str, Path]] = None,
        extra_data: Optional[Dict[str, Any]] = None,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        index_key: str = INDEX_KEY,
        sample_rate: int = DEFAULT_SAMPLE_RATE,
        max_num_seconds: Optional[float] = MAX_AUDIO_LEN,
        convert_to_mono: bool = DEFAULT_CONVERT_TO_MONO,
        start_times: Optional[List[float]] = None,
        spec_repr_func: Callable[[FloatTensor], FloatTensor] = default_spec_repr_func,
    ):
        """

        Args:
            paths: List of audio file paths.
            dataset_root: Base path for audio files.
            extra_data: Extra data to include in dataset items.
            input_tensors_key: Key under which audio tensors are stored.
            index_key: Key for indexing dataset items.
            sample_rate: Sampling rate of audio files.
            max_num_seconds: Duration to use for each audio file.
            convert_to_mono: Whether to downmix audio to one channel or leave the same.
            start_times: List of start time offsets in ``seconds`` for each audio.
            spec_repr_func: Spectral representation extraction function used for visualization.
        """
        assert (start_times is None) or (
            len(paths) == len(start_times)
        ), "The length of 'start_times' must match the length of 'paths' if 'start_times' is provided."
        assert sample_rate > 0, "The sample rate must be a positive integer."
        assert (
            max_num_seconds is None or max_num_seconds > 0
        ), "The maximum number of seconds must be None or a positive float."
        assert isinstance(convert_to_mono, bool), "'convert_to_mono' must be a boolean."

        paths = [Path(p) if dataset_root is None else Path(dataset_root) / p for p in paths]
        assert all(
            path.suffix in AUDIO_EXTENSIONS for path in paths
        ), f"Input audios should have one of '{AUDIO_EXTENSIONS}' extensions."

        if extra_data is not None:
            assert all(
                len(record) == len(paths) for record in extra_data.values()
            ), "All the extra records need to have the size equal to the dataset's size"
            self.extra_data = extra_data
        else:
            self.extra_data = {}

        self.input_tensors_key = input_tensors_key
        self.index_key = index_key

        self._paths = paths
        self._sample_rate = sample_rate

        if max_num_seconds is None:
            warnings.warn(
                "max_num_seconds is None, so the audio files will not be trimmed or padded. "
                "Additional collate_fn is required to handle different audio lengths."
            )
            self._num_frames = None
        else:
            self._num_frames = int(max_num_seconds * sample_rate)
        self._convert_to_mono = convert_to_mono
        self._frame_offsets = (
            [int(st * sample_rate) for st in start_times] if start_times is not None else [0] * len(paths)
        )
        self._spectral_function = spec_repr_func or default_spec_repr_func

    def _downmix_and_resample(self, audio: FloatTensor, sample_rate: int) -> FloatTensor:
        """
        (Optionally) downmix audio to mono and resample it to the dataset's sampling rate.

        Args:
            audio: Input audio tensor.
            sample_rate: Original sampling rate of the audio.

        Returns:
            Processed audio tensor.
        """
        from torchaudio.transforms import Resample

        if self._convert_to_mono and audio.shape[0] != 1:
            audio = audio.mean(dim=1, keepdim=True)
        if sample_rate != self._sample_rate:
            resampler = Resample(sample_rate, self._sample_rate)
            audio = resampler(audio)
        return audio

    @staticmethod
    def _trim_or_pad(audio: FloatTensor, frame_offset: int, num_frames: int) -> FloatTensor:
        """
        Trim and/or pad the audio to match the desired number of frames.

        Args:
            audio: Audio tensor.
            frame_offset: Frame offset for trimming the audio tensor.
            num_frames: Desired number of frames to be in the audio tensor.

        Returns:
            Trimmed and/or padded audio tensor.
        """
        if audio.shape[1] < frame_offset:
            raise ValueError(f"The frame offset {frame_offset} is greater than the audio length {audio.shape[1]}.")
        if audio.shape[1] > num_frames:
            audio = audio[:, frame_offset : frame_offset + num_frames]
        if audio.shape[1] < num_frames:
            padding = (num_frames - audio.shape[1], 0)
            audio = torch.nn.functional.pad(audio, padding)
        return audio

    def get_audio(self, item: int) -> FloatTensor:
        """
        Load and process an audio file.

        Args:
            item: Dataset item index.

        Returns:
            Processed audio tensor.
        """
        import torchaudio

        path = self._paths[item]
        audio, sample_rate = torchaudio.load(str(path))
        audio = self._downmix_and_resample(audio, sample_rate)
        if self._num_frames is not None:
            audio = self._trim_or_pad(audio, self._frame_offsets[item], self._num_frames)
        return audio

    def __getitem__(self, item: int) -> Dict[str, Union[FloatTensor, int]]:
        audio_tensor = self.get_audio(item)
        data = {
            self.input_tensors_key: audio_tensor,
            self.index_key: item,
        }
        for key, record in self.extra_data.items():
            if key in data:
                raise ValueError(f"<extra_data> and dataset share the same key: {key}")
            else:
                data[key] = record[item]
        return data

    def __len__(self) -> int:
        return len(self._paths)

    def visualize(self, item: int, color: TColor = BLACK) -> np.ndarray:
        """
        Visualize an audio file.

        Args:
            item: Dataset item index.
            color: Color of the plot.

        Returns:
            Array representing the image of the plot.
        """
        audio = self.get_audio(item)
        spec_repr = self._spectral_function(audio)
        return visualize_audio(spec_repr=spec_repr, color=color)

    def visualize_as_html(self, item: int, title: str, color: TColor = BLACK) -> str:
        """
        Visualize an audio file in HTML markup.

        Args:
            item: Dataset item index.
            color: Color of the plot.
            title: The title of html block.

        Returns:
            HTML markup with spectral representation image and audio player.
        """
        audio = self.get_audio(item)
        spec_repr = self._spectral_function(audio)
        html = visualize_audio_with_player(
            audio=audio, spec_repr=spec_repr, sample_rate=self._sample_rate, color=color, title=title
        )
        return html


class AudioLabeledDataset(DFLabeledDataset, IVisualizableDataset, IHTMLVisualizableDataset):
    """
    The dataset of audios having their ground truth labels.

    """

    _dataset: AudioBaseDataset

    def __init__(
        self,
        df: pd.DataFrame,
        dataset_root: Optional[Union[str, Path]] = None,
        extra_data: Optional[Dict[str, Any]] = None,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        index_key: str = INDEX_KEY,
        labels_key: str = LABELS_KEY,
        sample_rate: int = DEFAULT_SAMPLE_RATE,
        max_num_seconds: Optional[float] = MAX_AUDIO_LEN,
        convert_to_mono: bool = DEFAULT_CONVERT_TO_MONO,
        spec_repr_func: Callable[[FloatTensor], FloatTensor] = default_spec_repr_func,
    ):
        """

        Args:
            df: DataFrame with input data.
            dataset_root: Base path for audio files.
            extra_data: Extra data to include in dataset items.
            input_tensors_key: Key under which audio tensors are stored.
            index_key: Key for indexing dataset items.
            labels_key: Key under which labels are stored.
            sample_rate: Sampling rate of audio files.
            max_num_seconds: Duration to use from each audio file.
            convert_to_mono: Whether to downmix audio to one channel or leave the same.
            spec_repr_func: Spectral representation extraction function used for visualization.
        """
        dataset = AudioBaseDataset(
            paths=df[PATHS_COLUMN].tolist(),
            dataset_root=dataset_root,
            extra_data=extra_data,
            input_tensors_key=input_tensors_key,
            index_key=index_key,
            sample_rate=sample_rate,
            max_num_seconds=max_num_seconds,
            convert_to_mono=convert_to_mono,
            start_times=parse_start_times(df),
            spec_repr_func=spec_repr_func,
        )
        super().__init__(dataset=dataset, df=df, extra_data=extra_data, labels_key=labels_key)

    def visualize(self, item: int, color: TColor) -> np.ndarray:
        return self._dataset.visualize(item=item, color=color)

    def visualize_as_html(self, item: int, title: str, color: TColor) -> str:
        return self._dataset.visualize_as_html(item=item, title=title, color=color)


class AudioQueryGalleryDataset(DFQueryGalleryDataset, IVisualizableDataset, IHTMLVisualizableDataset):
    """
    The `non-annotated` dataset of audios having `query`/`gallery` split.
    To perform `1 vs rest` validation, where a query is evaluated versus the whole validation dataset
    (except for this exact query), you should mark the item as ``is_query == True`` and ``is_gallery == True``.

    """

    _dataset: AudioBaseDataset

    def __init__(
        self,
        df: pd.DataFrame,
        dataset_root: Optional[Union[str, Path]] = None,
        extra_data: Optional[Dict[str, Any]] = None,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        index_key: str = INDEX_KEY,
        sample_rate: int = DEFAULT_SAMPLE_RATE,
        max_num_seconds: Optional[float] = MAX_AUDIO_LEN,
        convert_to_mono: bool = DEFAULT_CONVERT_TO_MONO,
        spec_repr_func: Callable[[FloatTensor], FloatTensor] = default_spec_repr_func,
    ):
        """

        Args:
            df: DataFrame with input data.
            dataset_root: Base path for audio files.
            extra_data: Extra data to include in dataset items.
            input_tensors_key: Key under which audio tensors are stored.
            index_key: Key for indexing dataset items.
            sample_rate: Sampling rate of audio files.
            max_num_seconds: Duration to use from each audio file.
            convert_to_mono: Whether to downmix audio to one channel or leave the same.
            spec_repr_func: Spectral representation extraction function used for visualization.
        """
        dataset = AudioBaseDataset(
            paths=df[PATHS_COLUMN].tolist(),
            dataset_root=dataset_root,
            extra_data=extra_data,
            input_tensors_key=input_tensors_key,
            index_key=index_key,
            sample_rate=sample_rate,
            max_num_seconds=max_num_seconds,
            convert_to_mono=convert_to_mono,
            start_times=parse_start_times(df),
            spec_repr_func=spec_repr_func,
        )
        super().__init__(dataset=dataset, df=df, extra_data=extra_data)

    def visualize(self, item: int, color: TColor) -> np.ndarray:
        return self._dataset.visualize(item=item, color=color)

    def visualize_as_html(self, item: int, title: str, color: TColor) -> str:
        return self._dataset.visualize_as_html(item=item, color=color, title=title)


class AudioQueryGalleryLabeledDataset(DFQueryGalleryLabeledDataset, IVisualizableDataset, IHTMLVisualizableDataset):
    """
    The `annotated` dataset of audios having `query`/`gallery` split.
    To perform `1 vs rest` validation, where a query is evaluated versus the whole validation dataset
    (except for this exact query), you should mark the item as ``is_query == True`` and ``is_gallery == True``.

    """

    _dataset: AudioBaseDataset

    def __init__(
        self,
        df: pd.DataFrame,
        dataset_root: Optional[Union[str, Path]] = None,
        extra_data: Optional[Dict[str, Any]] = None,
        input_tensors_key: str = INPUT_TENSORS_KEY,
        index_key: str = INDEX_KEY,
        labels_key: str = LABELS_KEY,
        sample_rate: int = DEFAULT_SAMPLE_RATE,
        max_num_seconds: Optional[float] = MAX_AUDIO_LEN,
        convert_to_mono: bool = DEFAULT_CONVERT_TO_MONO,
        spec_repr_func: Callable[[FloatTensor], FloatTensor] = default_spec_repr_func,
    ):
        """

        Args:
            df: DataFrame with input data.
            dataset_root: Base path for audio files.
            extra_data: Extra data to include in dataset items.
            input_tensors_key: Key under which audio tensors are stored.
            index_key: Key for indexing dataset items.
            labels_key: Key under which labels are stored.
            sample_rate: Sampling rate of audio files.
            max_num_seconds: Duration to use from each audio file.
            convert_to_mono: Whether to downmix audio to one channel or leave the same.
            spec_repr_func: Spectral representation extraction function used for visualization.
        """
        dataset = AudioBaseDataset(
            paths=df[PATHS_COLUMN].tolist(),
            dataset_root=dataset_root,
            extra_data=extra_data,
            input_tensors_key=input_tensors_key,
            index_key=index_key,
            sample_rate=sample_rate,
            max_num_seconds=max_num_seconds,
            convert_to_mono=convert_to_mono,
            start_times=parse_start_times(df),
            spec_repr_func=spec_repr_func,
        )
        super().__init__(dataset=dataset, df=df, extra_data=extra_data, labels_key=labels_key)

    def visualize(self, item: int, color: TColor) -> np.ndarray:
        return self._dataset.visualize(item=item, color=color)

    def visualize_as_html(self, item: int, title: str, color: TColor) -> str:
        return self._dataset.visualize_as_html(item=item, color=color, title=title)


__all__ = [
    "AudioBaseDataset",
    "AudioLabeledDataset",
    "AudioQueryGalleryDataset",
    "AudioQueryGalleryLabeledDataset",
]



import base64
from io import BytesIO
from typing import Union

import matplotlib.pyplot as plt
import numpy as np
import torch
from PIL import Image
from torch import FloatTensor

from oml.const import BLACK, DEFAULT_SAMPLE_RATE, TColor


def default_spec_repr_func(audio: FloatTensor) -> FloatTensor:
    """
    Generate a default spectral representation (log-scaled MelSpec) from an audio signal.

    Args:
        audio: The input audio tensor.

    Returns:
        The spectral representation of the input audio tensor.
    """
    from torchaudio.transforms import MelSpectrogram

    melspectrogram = MelSpectrogram(sample_rate=DEFAULT_SAMPLE_RATE, n_fft=2048, hop_length=512, n_mels=128)
    melspec = melspectrogram(audio)
    log_melspec = torch.log1p(melspec).squeeze(0)
    return log_melspec


def _visualize_audio(
    spec_repr: FloatTensor, color: TColor = BLACK, draw_bbox: bool = True, return_b64: bool = False
) -> Union[np.ndarray, str]:
    """
    Internal function to visualize an audio spectrogram.

    Args:
        spec_repr: The spectrogram representation.
        color: The color of the bounding box.
        draw_bbox: Whether to draw a bounding box around the spectrogram.
        return_b64: Whether to return the image as a base64 string.

    Returns:
        If return_b64 is ``False``, returns the image as an array.
        If return_b64 is ``True``, returns the image as a base64 string.
    """
    fig, ax = plt.subplots(figsize=(2.56, 2.56), dpi=100)

    # actual image
    ax.imshow(spec_repr, aspect="auto", origin="lower")

    # bbox and axes
    if draw_bbox:
        frame_thickness = 5
        for axis in ["top", "bottom", "left", "right"]:
            ax.spines[axis].set_linewidth(frame_thickness)
            ax.spines[axis].set_edgecolor([c / 255 for c in color])
    ax.set_xticks([])
    ax.set_yticks([])

    # drawing
    fig.canvas.draw()
    buf = BytesIO()
    plt.savefig(buf, format="png", bbox_inches="tight")
    buf.seek(0)
    img: Union[np.ndarray, str] = (
        base64.b64encode(buf.getvalue()).decode("ascii") if return_b64 else np.array(Image.open(buf), dtype=np.uint8)
    )
    plt.close(fig)
    return img


def visualize_audio(spec_repr: FloatTensor, color: TColor = BLACK, draw_bbox: bool = True) -> np.ndarray:
    """
    Visualize an audio spectral representation and return it as a NumPy array.

    Args:
        spec_repr: The spectral representation.
        color: The color of the bounding box.
        draw_bbox: Whether to draw a bounding box around the spectral representation.

    Returns:
        The spectral representation image as an array.
    """
    return _visualize_audio(spec_repr, color, draw_bbox, return_b64=False)  # type: ignore


def visualize_audio_with_player(
    audio: FloatTensor,
    spec_repr: FloatTensor,
    sample_rate: int,
    title: str = "",
    color: TColor = BLACK,
    draw_bbox: bool = True,
) -> str:
    """
    Visualize an audio spectral representation and provide an HTML string with an audio player.

    Args:
        audio: The audio waveform.
        spec_repr: The spectral representation.
        sample_rate: The sampling rate of the audio.
        title: Title of the output HTML block if needed.
        color: The color of the bounding box.
        draw_bbox: Whether to draw a bounding box around the spectral representation.

    Returns:
        An HTML string that contains the spectral representation image and an audio player.
    """
    import torchaudio

    image_base64 = _visualize_audio(spec_repr, color, draw_bbox, return_b64=True)  # type: ignore

    buf = BytesIO()
    torchaudio.save(buf, audio, sample_rate=sample_rate, format="wav")
    buf.seek(0)
    audio_base64 = base64.b64encode(buf.getvalue()).decode("ascii")

    # generate HTML
    html = f"""
    <div style="margin: 10px; display: inline-block;">
        <div style="text-align: center; font-weight: bold; margin-bottom: 5px;">{title}</div>
        <div style="border:5px solid {color}; padding: 3px;">
            <img src="data:image/png;base64,{image_base64}" style="width:128px; height:128px;" alt="Mel Spectrogram" />
            <audio controls style="display: block; width: 100%; margin-top: 3px;">
                <source src="data:audio/wav;base64,{audio_base64}" type="audio/wav">
                Your browser does not support the audio element.
            </audio>
        </div>
    </div>
    """

    return html


__all__ = ["default_spec_repr_func", "visualize_audio", "visualize_audio_with_player"]



from typing import List, Union

import albumentations as albu
import cv2
from albumentations.pytorch import ToTensorV2

from oml.const import MEAN, MEAN_CLIP, PAD_COLOR, STD, STD_CLIP, TNormParam
from oml.utils.misc import adapt_argument_as_kwarg

TTransformsList = List[Union[albu.ImageOnlyTransform, albu.DualTransform]]


def get_normalisation_albu(mean: TNormParam = MEAN, std: TNormParam = STD) -> albu.Compose:
    return albu.Compose([albu.Normalize(mean=mean, std=std), ToTensorV2()])


def get_normalisation_resize_albu(im_size: int, mean: TNormParam = MEAN, std: TNormParam = STD) -> albu.Compose:
    pad_value_kwarg = adapt_argument_as_kwarg(albu.PadIfNeeded, ["value", "fill"], PAD_COLOR)
    return albu.Compose(
        [
            albu.LongestMaxSize(max_size=im_size),
            albu.PadIfNeeded(min_height=im_size, min_width=im_size, border_mode=cv2.BORDER_CONSTANT, **pad_value_kwarg),
            albu.Normalize(mean=mean, std=std),
            ToTensorV2(),
        ]
    )


def get_normalisation_resize_albu_clip(im_size: int) -> albu.Compose:
    return get_normalisation_resize_albu(im_size=im_size, mean=MEAN_CLIP, std=STD_CLIP)


__all__ = [
    "get_normalisation_albu",
    "get_normalisation_resize_albu",
    "get_normalisation_resize_albu_clip",
]



from pathlib import Path
from typing import Optional, Union

import torch

from oml.interfaces.models import IExtractor
from oml.models.audio.ecapa_tdnn.external.model import ECAPA_TDNN
from oml.models.utils import (
    remove_criterion_in_state_dict,
    remove_prefix_from_state_dict,
)
from oml.utils.io import download_checkpoint
from oml.utils.misc_torch import normalise


def ecapa_tdnn_taoruijie() -> ECAPA_TDNN:
    return ECAPA_TDNN(C=1024)


class ECAPATDNNExtractor(IExtractor):
    constructors = {
        "ecapa_tdnn_taoruijie": ecapa_tdnn_taoruijie,
    }
    pretrained_models = {
        "ecapa_tdnn_taoruijie": {
            "url": "https://github.com/TaoRuijie/ECAPA-TDNN/raw/refs/heads/main/exps/pretrain.model",
            "hash": "2bbc0ea09a41217eaff16b676ecb63d7",
            "fname": "ecapa_tdnn_taoruijie.pth",
            "init_args": {
                "arch": "ecapa_tdnn_taoruijie",
                "normalise_features": False,
            },
        }
    }

    def __init__(
        self,
        weights: Optional[Union[Path, str]],
        arch: str,
        normalise_features: bool = False,
    ):
        """
        Args:
            weights: Path to weights or special key for pretrained ones or ``None`` for random initialization.
             You can check available pretrained checkpoints in ``ECAPATDNNExtractor.pretrained_models``.
            arch: Model architecture, currently only supports ``ecapa_tdnn_taoruijie``.
            normalise_features: Set ``True`` to normalise output features.
        """
        super().__init__()

        assert arch in self.constructors, f"Unknown architecture: {arch}"

        self.normalise_features = normalise_features
        self.model = self.constructors[arch]()

        if weights is None:
            return

        if weights in self.pretrained_models:
            pretrained = self.pretrained_models[weights]  # type: ignore
            weights = download_checkpoint(
                pretrained["url"],  # type: ignore
                pretrained["hash"],  # type: ignore
                fname=pretrained["fname"],  # type: ignore
            )

        state_dict = torch.load(weights, map_location="cpu", weights_only=False)
        state_dict = state_dict["state_dict"] if "state_dict" in state_dict else state_dict
        state_dict = remove_criterion_in_state_dict(state_dict)
        ckpt = remove_prefix_from_state_dict(state_dict, trial_key="layer1.conv1.bias")

        self.model.load_state_dict(ckpt, strict=True)

    @property
    def feat_dim(self) -> int:
        return self.model.fc6.weight.shape[0]

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        assert x.ndim == 2 or (
            x.ndim == 3 and x.shape[1] == 1
        ), "The model expects input audio to have shape (batch_size, n_samples) or (batch_size, 1, n_samples)"

        if x.ndim == 3:
            x = x.squeeze(1)

        x = self.model.forward(x, aug=False)
        if self.normalise_features:
            x = normalise(x)

        return x


__all__ = ["ECAPATDNNExtractor"]
